/** 
	Copyright 2009 Bill Horne

	This file is part of the Poincare program for photographic
	tiling of the Poincare disk.

    Poincare is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Poincare is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Poincare.  If not, see <http://www.gnu.org/licenses/>.
**/

#include <cstdio>
#include <cstring>
#include <sys/stat.h>
#include <vector>
#include "Properties.h"

KeyValuePairList*
addKeyValuePairList(KeyValuePairList *head, char *key, char *value)
{
	KeyValuePairList *pair = new KeyValuePairList;
	pair->key = strdup(key);
	pair->value = strdup(value);
	pair->next = head;
	return pair;
}

typedef enum {START, KEY, ASSIGNMENT, VALUE, QUOTEDSTRING, UNQUOTEDSTRING, END, COMMENT} State;

typedef enum {WHITESPACE, SHARP, QUOTE, EQUALS, NEWLINE, ALPHA, OTHER} CharType;

void
throwParsingException(int lineno)
{
	char msg[1024];
	sprintf(msg, "parsing exception, line %d", lineno);
	throw msg;
}

CharType
getCharType(char c)
{
	switch (c) {
		case ' ':
		case '\r':
		case '\t':
			return WHITESPACE;
		case '#':
			return SHARP;
		case '"':
			return QUOTE;
		case '=':
			return EQUALS;
		case '\n':
			return NEWLINE;
		default:
			if ((c >= 'a') && (c <= 'z')) return ALPHA;
			if ((c >= 'A') && (c <= 'Z')) return ALPHA;
			if ((c >= ' ') && (c <= '~')) return OTHER;
	}
	throw "bad character";
}

// This is a hand generated parse from a finite state diagram.
KeyValuePairList*
parse(char *filename)
{
	FILE *fp;
	struct stat info;
	char *buf;
	char key[1024];
	char value[1024];

	KeyValuePairList *kvp = 0;

	if ((fp = fopen(filename, "rb")) == 0) {
		throw "can't open properties file";
	}

	if (stat(filename, &info)) {
		throw "stat properties file";
	}

	buf = new char[info.st_size];

	int r = fread(buf, sizeof(char), (int)info.st_size, fp);
	if (r != info.st_size) {
		throw "reading properties files";
	}

	char *p = buf;
	State state = START;
	int lineno = 1;
	int keyStart = 0;
	int valueStart = 0;
	
	for (int i=0; i<info.st_size; i++) {
		CharType type = getCharType(*p);

		if (state == START) {
			if (type == WHITESPACE)		{ state = START; } 
			else if (type == SHARP)		{ state = COMMENT; } 
			else if (type == NEWLINE)	{ state = START; }
			else if (type == ALPHA)		{
				state = KEY;
				keyStart = i;
			}
			else						{ throwParsingException(lineno); }
		
		} else if (state == KEY) {
			if (type == WHITESPACE)		{
				state = ASSIGNMENT;
				strncpy(key, &buf[keyStart], i - keyStart);
				key[i - keyStart] = 0;
			} 
			else if (type == EQUALS)	{
				state = VALUE;
				strncpy(key, &buf[keyStart], i - keyStart);
				key[i - keyStart] = 0;
				valueStart = i+1;
			} 
			else if (type == ALPHA)		{ state = KEY; }
			else						{ throwParsingException(lineno); }
		
		} else if (state == ASSIGNMENT) {
			if (type == WHITESPACE)		{ state = ASSIGNMENT; } 
			else if (type == EQUALS)	{
				state = VALUE;
				valueStart = i+1;
			} 
			else						{ throwParsingException(lineno); }
		
		} else if (state == VALUE) {
			if (type == WHITESPACE)		{ state = VALUE; } 
			else if (type == QUOTE)		{
				state = QUOTEDSTRING;
				valueStart = i+1;
			} 
			else if (type == ALPHA)		{
				state = UNQUOTEDSTRING;
				valueStart = i;
			}
			else if (type == OTHER)		{
				state = UNQUOTEDSTRING;
				valueStart = i;
			}
			else						{ throwParsingException(lineno); }
		} else if (state == QUOTEDSTRING) {
			if (type == WHITESPACE)		{ state = QUOTEDSTRING; } 
			else if (type == SHARP)		{ state = QUOTEDSTRING; } 
			else if (type == QUOTE)		{
				state = END;
				strncpy(value, &buf[valueStart], i - valueStart);
				value[i - valueStart] = 0;
				kvp = addKeyValuePairList(kvp, key, value);
			} 
			else if (type == EQUALS)	{ state = QUOTEDSTRING; } 
			else if (type == ALPHA)		{ state = QUOTEDSTRING; }
			else if (type == OTHER)		{ state = QUOTEDSTRING; }
			else						{ throwParsingException(lineno); }
		} else if (state == UNQUOTEDSTRING) {
			if (type == WHITESPACE)		{
				state = END;
				strncpy(value, &buf[valueStart], i - valueStart);
				value[i - valueStart] = 0;
				kvp = addKeyValuePairList(kvp, key, value);
			} 
			else if (type == NEWLINE)	{
				state = START;
				strncpy(value, &buf[valueStart], i - valueStart);
				value[i - valueStart] = 0;
				kvp = addKeyValuePairList(kvp, key, value);
			}
			else if (type == ALPHA)		{ state = UNQUOTEDSTRING; }
			else if (type == OTHER)		{ state = UNQUOTEDSTRING; }
			else						{ throwParsingException(lineno); }
		} else if (state == END) {
			if (type == WHITESPACE)		{ state = END; } 
			else if (type == SHARP)		{ state = COMMENT; } 
			else if (type == NEWLINE)	{ state = START; }
			else						{ throwParsingException(lineno); }
		} else if (state == COMMENT) {
			if (type == WHITESPACE)		{ state = COMMENT; } 
			else if (type == SHARP)		{ state = COMMENT; } 
			else if (type == QUOTE)		{ state = COMMENT; } 
			else if (type == EQUALS)	{ state = COMMENT; } 
			else if (type == NEWLINE)	{ state = START; }
			else if (type == ALPHA)		{ state = COMMENT; }
			else if (type == OTHER)		{ state = COMMENT; }
			else						{ throwParsingException(lineno); }
		} else {
			throwParsingException(lineno);
		}
		
		if (type == NEWLINE) lineno++;
		p++;
	}

	// Make sure we ended in a good state.
	if (state == KEY) {
		throwParsingException(lineno);
	} else if (state == ASSIGNMENT) {
		throwParsingException(lineno);
	} else if (state == VALUE) {
		throwParsingException(lineno);
	} else if (state == QUOTEDSTRING) {
		throwParsingException(lineno);
	}

	fclose(fp);

    return kvp;
}

Properties::Properties(char *filename)
{
	_kvp = parse(filename);
}

char*
Properties::getProperty(char *key)
{
	for (KeyValuePairList *p=_kvp; p!=0; p=p->next) {
		if (strcmp(p->key, key) == 0) {
			return p->value;
		}
	}
	return 0;
}
