/** 
	Copyright 2009 Bill Horne

	This file is part of the Poincare program for photographic
	tiling of the Poincare disk.

    Poincare is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Poincare is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Poincare.  If not, see <http://www.gnu.org/licenses/>.
**/

#include <cstdio>
#include <cstring>
#include <cmath>

#include "Circle.h"
#include "ImageProxy.h"
#include "AffineTransform.h"
#include "Point.h"
#include "Parameters.h"
#include "FundamentalRegion.h"


// TODO:  Better exception handling
int
main(int argc, char *argv[])
{
	try {
		if (argc != 2) {
			fprintf(stderr, "Usage: poincare parameter_file\n");
			return 1;
		}
		Parameters params(argv[1]);
		if ((params.p - 2)*(params.q - 2) <= 4) {
			throw "(p-2)(q-2) must be greater than 4";
		}

		Circle edge = createEdgeCircle(params.p, params.q);

		//*****************************************************
		//  MASK
		//*****************************************************
		if (strcmp(params.action, "mask") == 0) {
			double d = sqrt((params.py - params.ry)*(params.py - params.ry) + (params.px - params.rx)*(params.px - params.rx));
			double scale = (edge.center().x - edge.radius()) / d;
			double theta = -atan2(params.ry - params.py, params.rx - params.px);
			
			AffineTransform t;
			t.translate(params.px, params.py);
			t.rotate(theta);
			t.scale(scale);

			ImageProxy *inImage = new ImageProxy(params.infile);
			int width = inImage->getWidth();
			int height = inImage->getHeight();

			FundamentalRegion *region = new FundamentalRegion(params.p, params.q);;

			PixelProxy *maskColor = new PixelProxy(params.maskColor);

			for (int xidx=0; xidx<width; xidx++) {
				for (int yidx=0; yidx<height; yidx++) {
					Point in;
					in.x = (double)xidx;
					in.y = (double)yidx;

					Point a = t.transform(in);

					if (!region->inside(a)) {
						inImage->setPixel(xidx, yidx, maskColor);
					}
				}
			}

			inImage->write(params.outfile); 
		} 
		
		//*****************************************************
		//  RENDER
		//*****************************************************
		else if (strcmp(params.action, "render") == 0) {

			double d = sqrt((params.py - params.ry)*(params.py - params.ry) + (params.px - params.rx)*(params.px - params.rx));
			double scale = d / (edge.center().x - edge.radius());
			double theta = atan2(params.ry - params.py, params.rx - params.px);

			// Transform from hyperbolic space to input space
			AffineTransform hyp2in;
			hyp2in.scale(scale);
			hyp2in.rotate(theta);
			hyp2in.translate(-params.px, -params.py);

			// Transform from output space to hyperbolic space
			AffineTransform out2hyp;
			double iscale = 2.0 / (double)(params.renderedSize - 2*params.borderSize - 1);
			out2hyp.scale(iscale);
			// Strange things happen when the center is exactly on a pixel.  Perturb it slightly.
			// TODO:  Old problem.  Check if this is still true.
			double offset = -1.0 - iscale*params.borderSize + params.perturbationFactor;
			out2hyp.translate(-offset, -offset);

			ImageProxy *inImage = new ImageProxy(params.infile);
			int width = inImage->getWidth();
			int height = inImage->getHeight();

			ImageProxy *outImage = new ImageProxy(params.renderedSize, params.renderedSize);

			FundamentalRegion *region = new FundamentalRegion(params.p, params.q);

			Point origin;
			origin.x = 0.0;
			origin.y = 0.0;
			Circle unitCircle(origin, 1.0);

			PixelProxy *outOfBoundsColor = new PixelProxy(params.outOfBoundsColor);
			PixelProxy *unreachableColor = new PixelProxy(params.unreachableColor);
			PixelProxy *nonSourceColor = new PixelProxy(params.nonSourceColor);
			PixelProxy *renderedPixel = new PixelProxy(0, 0, 0);

			for (int xidx=0; xidx<params.renderedSize; xidx++) {
				for (int yidx=0; yidx<params.renderedSize; yidx++) {

					bool positive = true;
					
					bool allOutside = true;
					bool allOutOfBounds = true;
					int red=0;
					int green=0;
					int blue=0;
					int n=0;

					for (int xx=0; xx<params.supersamplingRate; xx++)  {
						for (int yy=0; yy<params.supersamplingRate; yy++)  {

							// transform (xidx,yidx) --> (x,y)
							Point a;
							a.x = (double)(xidx + xx*(1.0/(double)params.supersamplingRate));
							a.y = (double)(yidx + yy*(1.0/(double)params.supersamplingRate));
							a = out2hyp.transform(a);

							if (unitCircle.insideOrOn(a)) {
								allOutside = false;

								Point out;
								out.x = 0.0;
								out.y = 0.0;

								int index = 0;
								bool parity = false;
								bool err = region->reversePixelLookup(a, params.maxIterations, &out, &parity);
								if (err) {
									outImage->setPixel(xidx, yidx, unreachableColor);
									//printf("maxIterations exceeded\n");
								}

								bool negated = false;
								if (params.negate && parity) {
									negated = true;
								}

								// Find corresponding input pixel
								Point b = hyp2in.transform(out);

								PixelProxy *pp = inImage->getSubPixel(b.x, b.y);
								if (pp != 0) {
									allOutOfBounds = false;
									if (negated) {
										pp->negate();
									}
									red += pp->red();
									green += pp->green();
									blue += pp->blue();
									n++;
								}
								delete pp;
							}
						}
					}

					if (allOutside) {
						outImage->setPixel(xidx, yidx, outOfBoundsColor);
					} else if (allOutOfBounds) {
						outImage->setPixel(xidx, yidx, nonSourceColor);
					} else {
						renderedPixel->setRed((int)((double)red/(double)n + 0.5));
						renderedPixel->setGreen((int)((double)green/(double)n + 0.5));
						renderedPixel->setBlue((int)((double)blue/(double)n + 0.5));
						outImage->setPixel(xidx, yidx, renderedPixel);
					}
				}
			}

			outImage->write(params.outfile); 
		}
	} catch (const char* e) {
		fprintf(stderr, "Exception: %s\n", e);
		return 1;
	}

	return 0;
}